@@ -14,12 +14,12 @@
 import collections
 import logging
 import re
+import sys
 
+from sqlalchemy import event
 from sqlalchemy import exc as sqla_exc
 
-from oslo_db._i18n import _LE
 from oslo_db import exception
-from oslo_db.sqlalchemy import compat
 
 
 LOG = logging.getLogger(__name__)
@@ -58,9 +58,13 @@ def filters(dbname, exception_type, regex):
 # as well as sqlalchemy.exc.DBAPIError, as SQLAlchemy will reraise it
 # as this until issue #3075 is fixed.
 @filters("mysql", sqla_exc.OperationalError, r"^.*\b1213\b.*Deadlock found.*")
-@filters("mysql", sqla_exc.OperationalError,
+@filters("mysql", sqla_exc.DatabaseError,
          r"^.*\b1205\b.*Lock wait timeout exceeded.*")
 @filters("mysql", sqla_exc.InternalError, r"^.*\b1213\b.*Deadlock found.*")
+@filters("mysql", sqla_exc.InternalError,
+         r"^.*\b1213\b.*detected deadlock/conflict.*")
+@filters("mysql", sqla_exc.InternalError,
+         r"^.*\b1213\b.*Deadlock: wsrep aborted.*")
 @filters("postgresql", sqla_exc.OperationalError, r"^.*deadlock detected.*")
 @filters("postgresql", sqla_exc.DBAPIError, r"^.*deadlock detected.*")
 @filters("ibm_db_sa", sqla_exc.DBAPIError, r"^.*SQL0911N.*")
@@ -91,11 +95,11 @@ def _deadlock_error(operational_error, match, engine_name, is_disconnect):
 
 
 @filters("mysql", sqla_exc.IntegrityError,
-         r"^.*\b1062\b.*Duplicate entry '(?P<value>.+)'"
+         r"^.*\b1062\b.*Duplicate entry '(?P<value>.*)'"
          r" for key '(?P<columns>[^']+)'.*$")
 # NOTE(jd) For binary types
 @filters("mysql", sqla_exc.IntegrityError,
-         r"^.*\b1062\b.*Duplicate entry \\'(?P<value>.+)\\'"
+         r"^.*\b1062\b.*Duplicate entry \\'(?P<value>.*)\\'"
          r" for key \\'(?P<columns>.+)\\'.*$")
 # NOTE(pkholkin): the first regex is suitable only for PostgreSQL 9.x versions
 #                 the second regex is suitable for PostgreSQL 8.x versions
@@ -191,15 +195,16 @@ def _sqlite_dupe_key_error(integrity_error, match, engine_name, is_disconnect):
          r"(?i).*foreign key constraint failed")
 @filters("postgresql", sqla_exc.IntegrityError,
          r".*on table \"(?P<table>[^\"]+)\" violates "
-         "foreign key constraint \"(?P<constraint>[^\"]+)\".*\n"
-         "DETAIL:  Key \((?P<key>.+)\)=\(.+\) "
-         "is (not present in|still referenced from) table "
-         "\"(?P<key_table>[^\"]+)\".")
-@filters("mysql", sqla_exc.IntegrityError,
-         r".* u?'Cannot (add|delete) or update a (child|parent) row: "
-         'a foreign key constraint fails \([`"].+[`"]\.[`"](?P<table>.+)[`"], '
-         'CONSTRAINT [`"](?P<constraint>.+)[`"] FOREIGN KEY '
-         '\([`"](?P<key>.+)[`"]\) REFERENCES [`"](?P<key_table>.+)[`"] ')
+         r"foreign key constraint \"(?P<constraint>[^\"]+)\".*\n"
+         r"DETAIL:  Key \((?P<key>.+)\)=\(.+\) "
+         r"is (not present in|still referenced from) table "
+         r"\"(?P<key_table>[^\"]+)\".")
+@filters(
+    "mysql", sqla_exc.IntegrityError,
+    r".*Cannot (add|delete) or update a (child|parent) row: "
+    r'a foreign key constraint fails \([`"].+[`"]\.[`"](?P<table>.+)[`"], '
+    r'CONSTRAINT [`"](?P<constraint>.+)[`"] FOREIGN KEY '
+    r'\([`"](?P<key>.+)[`"]\) REFERENCES [`"](?P<key_table>.+)[`"] ')
 def _foreign_key_error(integrity_error, match, engine_name, is_disconnect):
     """Filter for foreign key errors."""
 
@@ -244,6 +249,69 @@ def _check_constraint_error(
     raise exception.DBConstraintError(table, check_name, integrity_error)
 
 
+@filters("postgresql", sqla_exc.ProgrammingError,
+         r".* constraint \"(?P<constraint>.+)\" "
+         "of relation "
+         "\"(?P<relation>.+)\" does not exist")
+@filters("mysql", sqla_exc.InternalError,
+         r".*1091,.*Can't DROP (?:FOREIGN KEY )?['`](?P<constraint>.+)['`]; "
+         "check that .* exists")
+@filters("mysql", sqla_exc.OperationalError,
+         r".*1091,.*Can't DROP (?:FOREIGN KEY )?['`](?P<constraint>.+)['`]; "
+         "check that .* exists")
+@filters("mysql", sqla_exc.InternalError,
+         r".*1025,.*Error on rename of '.+/(?P<relation>.+)' to ")
+def _check_constraint_non_existing(
+        programming_error, match, engine_name, is_disconnect):
+    """Filter for constraint non existing errors."""
+
+    try:
+        relation = match.group("relation")
+    except IndexError:
+        relation = None
+
+    try:
+        constraint = match.group("constraint")
+    except IndexError:
+        constraint = None
+
+    raise exception.DBNonExistentConstraint(relation,
+                                            constraint,
+                                            programming_error)
+
+
+@filters("sqlite", sqla_exc.OperationalError,
+         r".* no such table: (?P<table>.+)")
+@filters("mysql", sqla_exc.InternalError,
+         r".*1051,.*Unknown table '(.+\.)?(?P<table>.+)'\"")
+@filters("mysql", sqla_exc.OperationalError,
+         r".*1051,.*Unknown table '(.+\.)?(?P<table>.+)'\"")
+@filters("postgresql", sqla_exc.ProgrammingError,
+         r".* table \"(?P<table>.+)\" does not exist")
+def _check_table_non_existing(
+        programming_error, match, engine_name, is_disconnect):
+    """Filter for table non existing errors."""
+    raise exception.DBNonExistentTable(match.group("table"), programming_error)
+
+
+@filters("mysql", sqla_exc.InternalError,
+         r".*1049,.*Unknown database '(?P<database>.+)'\"")
+@filters("mysql", sqla_exc.OperationalError,
+         r".*1049,.*Unknown database '(?P<database>.+)'\"")
+@filters("postgresql", sqla_exc.OperationalError,
+         r".*database \"(?P<database>.+)\" does not exist")
+@filters("sqlite", sqla_exc.OperationalError,
+         ".*unable to open database file.*")
+def _check_database_non_existing(
+        error, match, engine_name, is_disconnect):
+    try:
+        database = match.group("database")
+    except IndexError:
+        database = None
+
+    raise exception.DBNonExistentDatabase(database, error)
+
+
 @filters("ibm_db_sa", sqla_exc.IntegrityError, r"^.*SQL0803N.*$")
 def _db2_dupe_key_error(integrity_error, match, engine_name, is_disconnect):
     """Filter for DB2 duplicate key errors.
@@ -280,12 +348,26 @@ def _raise_mysql_table_doesnt_exist_asis(
          r".*1265.*Data truncated for column.*")
 @filters("mysql", sqla_exc.DataError,
          r".*1264.*Out of range value for column.*")
+@filters("mysql", sqla_exc.InternalError,
+         r"^.*1366.*Incorrect string value:*")
+@filters("sqlite", sqla_exc.ProgrammingError,
+         r"(?i).*You must not use 8-bit bytestrings*")
+@filters("mysql", sqla_exc.DataError,
+         r".*1406.*Data too long for column.*")
 def _raise_data_error(error, match, engine_name, is_disconnect):
     """Raise DBDataError exception for different data errors."""
 
     raise exception.DBDataError(error)
 
 
+@filters("mysql", sqla_exc.OperationalError,
+         r".*\(1305,\s+\'SAVEPOINT\s+(.+)\s+does not exist\'\)")
+def _raise_savepoints_as_dberrors(error, match, engine_name, is_disconnect):
+    # NOTE(rpodolyaka): this is a special case of an OperationalError that used
+    # to be an InternalError. It's expected to be wrapped into oslo.db error.
+    raise exception.DBError(error)
+
+
 @filters("*", sqla_exc.OperationalError, r".*")
 def _raise_operational_errors_directly_filter(operational_error,
                                               match, engine_name,
@@ -305,7 +387,9 @@ def _raise_operational_errors_directly_filter(operational_error,
         raise operational_error
 
 
-@filters("mysql", sqla_exc.OperationalError, r".*\(.*(?:2002|2003|2006|2013)")
+@filters("mysql", sqla_exc.OperationalError, r".*\(.*(?:2002|2003|2006|2013|1047)")  # noqa
+@filters("mysql", sqla_exc.InternalError, r".*\(.*(?:1927)")  # noqa
+@filters("mysql", sqla_exc.InternalError, r".*Packet sequence number wrong")  # noqa
 @filters("postgresql", sqla_exc.OperationalError, r".*could not connect to server")  # noqa
 @filters("ibm_db_sa", sqla_exc.OperationalError, r".*(?:30081)")
 def _is_db_connection_error(operational_error, match, engine_name,
@@ -314,6 +398,11 @@ def _is_db_connection_error(operational_error, match, engine_name,
     raise exception.DBConnectionError(operational_error)
 
 
+@filters("*", sqla_exc.NotSupportedError, r".*")
+def _raise_for_NotSupportedError(error, match, engine_name, is_disconnect):
+    raise exception.DBNotSupportedError(error)
+
+
 @filters("*", sqla_exc.DBAPIError, r".*")
 def _raise_for_remaining_DBAPIError(error, match, engine_name, is_disconnect):
     """Filter for remaining DBAPIErrors.
@@ -324,8 +413,7 @@ def _raise_for_remaining_DBAPIError(error, match, engine_name, is_disconnect):
     if is_disconnect:
         raise exception.DBConnectionError(error)
     else:
-        LOG.exception(
-            _LE('DBAPIError exception wrapped from %s') % error)
+        LOG.warn('DBAPIError exception wrapped.', exc_info=True)
         raise exception.DBError(error)
 
 
@@ -336,9 +424,11 @@ def _raise_for_unicode_encode(error, match, engine_name, is_disconnect):
 
 @filters("*", Exception, r".*")
 def _raise_for_all_others(error, match, engine_name, is_disconnect):
-    LOG.exception(_LE('DB exception wrapped.'))
+    LOG.warn('DB exception wrapped.', exc_info=True)
     raise exception.DBError(error)
 
+ROLLBACK_CAUSE_KEY = 'oslo.db.sp_rollback_cause'
+
 
 def handler(context):
     """Iterate through available filters and invoke those which match.
@@ -372,20 +462,57 @@ def handler(context):
                                     match,
                                     context.engine.dialect.name,
                                     context.is_disconnect)
-                            except exception.DBConnectionError:
-                                context.is_disconnect = True
-                                raise
+                            except exception.DBError as dbe:
+                                if (
+                                    context.connection is not None and
+                                    not context.connection.closed and
+                                    not context.connection.invalidated and
+                                    ROLLBACK_CAUSE_KEY
+                                    in context.connection.info
+                                ):
+                                    dbe.cause = \
+                                        context.connection.info.pop(
+                                            ROLLBACK_CAUSE_KEY)
+
+                                if isinstance(
+                                        dbe, exception.DBConnectionError):
+                                    context.is_disconnect = True
+                                return dbe
 
 
 def register_engine(engine):
-    compat.handle_error(engine, handler)
+    event.listen(engine, "handle_error", handler, retval=True)
 
+    @event.listens_for(engine, "rollback_savepoint")
+    def rollback_savepoint(conn, name, context):
+        exc_info = sys.exc_info()
+        if exc_info[1]:
+            # NOTE(zzzeek) accessing conn.info on an invalidated
+            # connection causes it to reconnect, which we don't
+            # want to do inside a rollback handler
+            if not conn.invalidated:
+                conn.info[ROLLBACK_CAUSE_KEY] = exc_info[1]
+        # NOTE(zzzeek) this eliminates a reference cycle between tracebacks
+        # that would occur in Python 3 only, which has been shown to occur if
+        # this function were in fact part of the traceback.  That's not the
+        # case here however this is left as a defensive measure.
+        del exc_info
 
-def handle_connect_error(engine):
-    """Connect to the engine, including handle_error handlers.
+    # try to clear the "cause" ASAP outside of savepoints,
+    # by grabbing the end of transaction events...
+    @event.listens_for(engine, "rollback")
+    @event.listens_for(engine, "commit")
+    def pop_exc_tx(conn):
+        # NOTE(zzzeek) accessing conn.info on an invalidated
+        # connection causes it to reconnect, which we don't
+        # want to do inside a rollback handler
+        if not conn.invalidated:
+            conn.info.pop(ROLLBACK_CAUSE_KEY, None)
 
-    The compat library now builds this into the engine.connect()
-    system as per SQLAlchemy 1.0's behavior.
-
-    """
-    return engine.connect()
+    # .. as well as connection pool checkin (just in case).
+    # the .info dictionary lasts as long as the DBAPI connection itself
+    # and is cleared out when the connection is recycled or closed
+    # due to invalidate etc.
+    @event.listens_for(engine, "checkin")
+    def pop_exc_checkin(dbapi_conn, connection_record):
+        connection_record.info.pop(ROLLBACK_CAUSE_KEY, None)
