@@ -17,12 +17,16 @@
 #    under the License.
 
 import collections
+from collections import abc
 import contextlib
+import inspect as pyinspect
+import itertools
 import logging
 import re
 
+from alembic.migration import MigrationContext
+from alembic.operations import Operations
 from oslo_utils import timeutils
-import six
 import sqlalchemy
 from sqlalchemy import Boolean
 from sqlalchemy import CheckConstraint
@@ -30,20 +34,23 @@ from sqlalchemy import Column
 from sqlalchemy.engine import Connectable
 from sqlalchemy.engine import reflection
 from sqlalchemy.engine import url as sa_url
+from sqlalchemy import exc
 from sqlalchemy import func
 from sqlalchemy import Index
 from sqlalchemy import inspect
 from sqlalchemy import Integer
 from sqlalchemy import MetaData
+from sqlalchemy.sql.expression import cast
 from sqlalchemy.sql.expression import literal_column
 from sqlalchemy.sql import text
 from sqlalchemy import String
 from sqlalchemy import Table
 from sqlalchemy.types import NullType
 
+from oslo_db._i18n import _
 from oslo_db import exception
-from oslo_db._i18n import _, _LI, _LW
 from oslo_db.sqlalchemy import models
+from oslo_db.sqlalchemy import ndb
 
 # NOTE(ochuprykov): Add references for backwards compatibility
 InvalidSortKey = exception.InvalidSortKey
@@ -53,30 +60,9 @@ LOG = logging.getLogger(__name__)
 
 _DBURL_REGEX = re.compile(r"[^:]+://([^:]+):([^@]+)@.+")
 
-
-def get_callable_name(function):
-    # TODO(harlowja): Replace this once
-    # it is possible to use https://review.openstack.org/#/c/122495/ which is
-    # a more complete and expansive module that does a similar thing...
-    try:
-        method_self = six.get_method_self(function)
-    except AttributeError:
-        method_self = None
-    if method_self is not None:
-        if isinstance(method_self, six.class_types):
-            im_class = method_self
-        else:
-            im_class = type(method_self)
-        try:
-            parts = (im_class.__module__, function.__qualname__)
-        except AttributeError:
-            parts = (im_class.__module__, im_class.__name__, function.__name__)
-    else:
-        try:
-            parts = (function.__module__, function.__qualname__)
-        except AttributeError:
-            parts = (function.__module__, function.__name__)
-    return '.'.join(parts)
+_VALID_SORT_DIR = [
+    "-".join(x) for x in itertools.product(["asc", "desc"],
+                                           ["nullsfirst", "nullslast"])]
 
 
 def sanitize_db_url(url):
@@ -86,6 +72,71 @@ def sanitize_db_url(url):
     return url
 
 
+def get_unique_keys(model):
+    """Get a list of sets of unique model keys.
+
+    :param model: the ORM model class
+    :rtype: list of sets of strings
+    :return: unique model keys or None if unable to find them
+    """
+
+    try:
+        mapper = inspect(model)
+    except exc.NoInspectionAvailable:
+        return None
+    else:
+        local_table = mapper.local_table
+        base_table = mapper.base_mapper.local_table
+
+        if local_table is None:
+            return None
+
+    # extract result from cache if present
+    has_info = hasattr(local_table, 'info')
+    if has_info:
+        info = local_table.info
+        if 'oslodb_unique_keys' in info:
+            return info['oslodb_unique_keys']
+
+    res = []
+    try:
+        constraints = base_table.constraints
+    except AttributeError:
+        constraints = []
+    for constraint in constraints:
+        # filter out any CheckConstraints
+        if isinstance(constraint, (sqlalchemy.UniqueConstraint,
+                                   sqlalchemy.PrimaryKeyConstraint)):
+            res.append({c.name for c in constraint.columns})
+    try:
+        indexes = base_table.indexes
+    except AttributeError:
+        indexes = []
+    for index in indexes:
+        if index.unique:
+            res.append({c.name for c in index.columns})
+    # cache result for next calls with the same model
+    if has_info:
+        info['oslodb_unique_keys'] = res
+    return res
+
+
+def _stable_sorting_order(model, sort_keys):
+    """Check whether the sorting order is stable.
+
+    :return: True if it is stable, False if it's not, None if it's impossible
+    to determine.
+    """
+    keys = get_unique_keys(model)
+    if keys is None:
+        return None
+    sort_keys_set = set(sort_keys)
+    for unique_keys in keys:
+        if unique_keys.issubset(sort_keys_set):
+            return True
+    return False
+
+
 # copy from glance/db/sqlalchemy/api.py
 def paginate_query(query, model, limit, sort_keys, marker=None,
                    sort_dir=None, sort_dirs=None):
@@ -100,12 +151,18 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
     the lexicographical ordering:
     (k1 > X1) or (k1 == X1 && k2 > X2) or (k1 == X1 && k2 == X2 && k3 > X3)
 
-    We also have to cope with different sort_directions.
+    We also have to cope with different sort_directions and cases where k2,
+    k3, ... are nullable.
 
     Typically, the id of the last row is used as the client-facing pagination
     marker, then the actual marker object must be fetched from the db and
     passed in to us as marker.
 
+    The "offset" parameter is intentionally avoided. As offset requires a
+    full scan through the preceding results each time, criteria-based
+    pagination is preferred. See http://use-the-index-luke.com/no-offset
+    for further background.
+
     :param query: the query object to which we should add paging/sorting
     :param model: the ORM model class
     :param limit: maximum number of items to return
@@ -113,18 +170,20 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
     :param marker: the last item of the previous page; we returns the next
                     results after this value.
     :param sort_dir: direction in which results should be sorted (asc, desc)
+                     suffix -nullsfirst, -nullslast can be added to defined
+                     the ordering of null values
     :param sort_dirs: per-column array of sort_dirs, corresponding to sort_keys
 
     :rtype: sqlalchemy.orm.query.Query
     :return: The query with sorting/pagination added.
     """
+    if _stable_sorting_order(model, sort_keys) is False:
+        LOG.warning('Unique keys not in sort_keys. '
+                    'The sorting order may be unstable.')
 
-    if 'id' not in sort_keys:
-        # TODO(justinsb): If this ever gives a false-positive, check
-        # the actual primary key, rather than assuming its id
-        LOG.warning(_LW('Id not in sort_keys; is sort_keys unique?'))
-
-    assert(not (sort_dir and sort_dirs))
+    if sort_dir and sort_dirs:
+        raise AssertionError('Disallow set sort_dir and '
+                             'sort_dirs at the same time.')
 
     # Default the sort direction to ascending
     if sort_dirs is None and sort_dir is None:
@@ -134,26 +193,37 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
     if sort_dirs is None:
         sort_dirs = [sort_dir for _sort_key in sort_keys]
 
-    assert(len(sort_dirs) == len(sort_keys))
+    if len(sort_dirs) != len(sort_keys):
+        raise AssertionError('sort_dirs and sort_keys must have same length.')
 
     # Add sorting
     for current_sort_key, current_sort_dir in zip(sort_keys, sort_dirs):
         try:
-            sort_dir_func = {
-                'asc': sqlalchemy.asc,
-                'desc': sqlalchemy.desc,
-            }[current_sort_dir]
+            inspect(model).all_orm_descriptors[current_sort_key]
         except KeyError:
-            raise ValueError(_("Unknown sort direction, "
-                               "must be 'desc' or 'asc'"))
-        try:
-            inspect(model).\
-                all_orm_descriptors[current_sort_key]
-        except KeyError:
-            raise exception.InvalidSortKey()
+            raise exception.InvalidSortKey(current_sort_key)
         else:
             sort_key_attr = getattr(model, current_sort_key)
 
+        try:
+            main_sort_dir, __, null_sort_dir = current_sort_dir.partition("-")
+            sort_dir_func = {
+                'asc': sqlalchemy.asc,
+                'desc': sqlalchemy.desc,
+            }[main_sort_dir]
+
+            null_order_by_stmt = {
+                "": None,
+                "nullsfirst": sort_key_attr.is_(None),
+                "nullslast": sort_key_attr.isnot(None),
+            }[null_sort_dir]
+        except KeyError:
+            raise ValueError(_("Unknown sort direction, "
+                               "must be one of: %s") %
+                             ", ".join(_VALID_SORT_DIR))
+
+        if null_order_by_stmt is not None:
+            query = query.order_by(sqlalchemy.desc(null_order_by_stmt))
         query = query.order_by(sort_dir_func(sort_key_attr))
 
     # Add pagination
@@ -167,18 +237,36 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
         criteria_list = []
         for i in range(len(sort_keys)):
             crit_attrs = []
-            for j in range(i):
-                model_attr = getattr(model, sort_keys[j])
-                crit_attrs.append((model_attr == marker_values[j]))
+            # NOTE: We skip the marker value comparison if marker_values[i] is
+            #       None, for two reasons: 1) the comparison operators below
+            #       ('<', '>') are not applicable on None value; 2) this is
+            #       safe because we can assume the primary key is included in
+            #       sort_key, thus checked as (one of) marker values.
+            if marker_values[i] is not None:
+                for j in range(i):
+                    model_attr = getattr(model, sort_keys[j])
+                    if marker_values[j] is not None:
+                        crit_attrs.append((model_attr == marker_values[j]))
 
-            model_attr = getattr(model, sort_keys[i])
-            if sort_dirs[i] == 'desc':
-                crit_attrs.append((model_attr < marker_values[i]))
-            else:
-                crit_attrs.append((model_attr > marker_values[i]))
-
-            criteria = sqlalchemy.sql.and_(*crit_attrs)
-            criteria_list.append(criteria)
+                model_attr = getattr(model, sort_keys[i])
+                val = marker_values[i]
+                # sqlalchemy doesn't like booleans in < >. bug/1656947
+                if isinstance(model_attr.type, Boolean):
+                    val = int(val)
+                    model_attr = cast(model_attr, Integer)
+                if sort_dirs[i].startswith('desc'):
+                    crit_attr = (model_attr < val)
+                    if sort_dirs[i].endswith('nullsfirst'):
+                        crit_attr = sqlalchemy.sql.or_(crit_attr,
+                                                       model_attr.is_(None))
+                else:
+                    crit_attr = (model_attr > val)
+                    if sort_dirs[i].endswith('nullslast'):
+                        crit_attr = sqlalchemy.sql.or_(crit_attr,
+                                                       model_attr.is_(None))
+                crit_attrs.append(crit_attr)
+                criteria = sqlalchemy.sql.and_(*crit_attrs)
+                criteria_list.append(criteria)
 
         f = sqlalchemy.sql.or_(*criteria_list)
         query = query.filter(f)
@@ -192,8 +280,7 @@ def paginate_query(query, model, limit, sort_keys, marker=None,
 def to_list(x, default=None):
     if x is None:
         return default
-    if not isinstance(x, collections.Iterable) or \
-            isinstance(x, six.string_types):
+    if not isinstance(x, abc.Iterable) or isinstance(x, str):
         return [x]
     elif isinstance(x, list):
         return x
@@ -359,42 +446,6 @@ def get_table(engine, name):
     return Table(name, metadata, autoload=True)
 
 
-class InsertFromSelect(object):
-    """Form the base for `INSERT INTO table (SELECT ... )` statement.
-
-    DEPRECATED: this class is deprecated and will be removed from oslo.db
-    in a few releases. Use default SQLAlchemy insert from select implementation
-    instead
-
-    :param table: table to insert records
-    :param select: select query
-    :param cols: list of columns to specify in insert clause
-    :return: SQLAlchemy :class:`Insert` object instance
-
-    Usage:
-
-    .. code-block:: python
-
-      select = sql.select(table_from)
-      insert = InsertFromSelect(table_to, select,
-                                ['id', 'name', 'insert_date'])
-      engine.execute(insert)
-
-    """
-    # NOTE(tdurakov): Insert from select implementation added to SQLAlchemy
-    # starting from version 0.8.7. Default SQLAlchemy implementation should be
-    # used instead of this. Deprecated.
-
-    def __new__(cls, table, select, cols=None):
-        if not cols:
-            cols = [c.name for c in table.c]
-
-        return table.insert(inline=True).from_select(cols, select)
-
-    def __init__(self, table, select, cols=None):
-        pass
-
-
 def _get_not_supported_column(col_name_col_instance, column_name):
     try:
         column = col_name_col_instance[column_name]
@@ -412,21 +463,21 @@ def _get_not_supported_column(col_name_col_instance, column_name):
     return column
 
 
-def drop_old_duplicate_entries_from_table(migrate_engine, table_name,
+def drop_old_duplicate_entries_from_table(engine, table_name,
                                           use_soft_delete, *uc_column_names):
     """Drop all old rows having the same values for columns in uc_columns.
 
     This method drop (or mark ad `deleted` if use_soft_delete is True) old
     duplicate rows form table with name `table_name`.
 
-    :param migrate_engine:  Sqlalchemy engine
+    :param engine:          Sqlalchemy engine
     :param table_name:      Table with duplicates
     :param use_soft_delete: If True - values will be marked as `deleted`,
                             if False - values will be removed from table
     :param uc_column_names: Unique constraint columns
     """
     meta = MetaData()
-    meta.bind = migrate_engine
+    meta.bind = engine
 
     table = Table(table_name, meta, autoload=True)
     columns_for_group_by = [table.c[name] for name in uc_column_names]
@@ -438,7 +489,7 @@ def drop_old_duplicate_entries_from_table(migrate_engine, table_name,
         columns_for_select, group_by=columns_for_group_by,
         having=func.count(table.c.id) > 1)
 
-    for row in migrate_engine.execute(duplicated_rows_select).fetchall():
+    for row in engine.execute(duplicated_rows_select).fetchall():
         # NOTE(boris-42): Do not remove row that has the biggest ID.
         delete_condition = table.c.id != row[0]
         is_none = None  # workaround for pyflakes
@@ -448,9 +499,9 @@ def drop_old_duplicate_entries_from_table(migrate_engine, table_name,
 
         rows_to_delete_select = sqlalchemy.sql.select(
             [table.c.id]).where(delete_condition)
-        for row in migrate_engine.execute(rows_to_delete_select).fetchall():
-            LOG.info(_LI("Deleting duplicated row with id: %(id)s from table: "
-                         "%(table)s"), dict(id=row[0], table=table_name))
+        for row in engine.execute(rows_to_delete_select).fetchall():
+            LOG.info("Deleting duplicated row with id: %(id)s from table: "
+                     "%(table)s", dict(id=row[0], table=table_name))
 
         if use_soft_delete:
             delete_statement = table.update().\
@@ -462,7 +513,7 @@ def drop_old_duplicate_entries_from_table(migrate_engine, table_name,
                 })
         else:
             delete_statement = table.delete().where(delete_condition)
-        migrate_engine.execute(delete_statement)
+        engine.execute(delete_statement)
 
 
 def _get_default_deleted_value(table):
@@ -473,11 +524,10 @@ def _get_default_deleted_value(table):
     raise exception.ColumnError(_("Unsupported id columns type"))
 
 
-def _restore_indexes_on_deleted_columns(migrate_engine, table_name, indexes):
-    table = get_table(migrate_engine, table_name)
+def _restore_indexes_on_deleted_columns(engine, table_name, indexes):
+    table = get_table(engine, table_name)
 
-    insp = reflection.Inspector.from_engine(migrate_engine)
-    real_indexes = insp.get_indexes(table_name)
+    real_indexes = get_indexes(engine, table_name)
     existing_index_names = dict(
         [(index['name'], index['column_names']) for index in real_indexes])
 
@@ -489,22 +539,20 @@ def _restore_indexes_on_deleted_columns(migrate_engine, table_name, indexes):
         if name in existing_index_names:
             column_names = [table.c[c] for c in existing_index_names[name]]
             old_index = Index(name, *column_names, unique=index["unique"])
-            old_index.drop(migrate_engine)
+            old_index.drop(engine)
 
         column_names = [table.c[c] for c in index['column_names']]
         new_index = Index(index["name"], *column_names, unique=index["unique"])
-        new_index.create(migrate_engine)
+        new_index.create(engine)
 
 
-def change_deleted_column_type_to_boolean(migrate_engine, table_name,
+def change_deleted_column_type_to_boolean(engine, table_name,
                                           **col_name_col_instance):
-    if migrate_engine.name == "sqlite":
+    if engine.name == "sqlite":
         return _change_deleted_column_type_to_boolean_sqlite(
-            migrate_engine, table_name, **col_name_col_instance)
-    insp = reflection.Inspector.from_engine(migrate_engine)
-    indexes = insp.get_indexes(table_name)
-
-    table = get_table(migrate_engine, table_name)
+            engine, table_name, **col_name_col_instance)
+    indexes = get_indexes(engine, table_name)
+    table = get_table(engine, table_name)
 
     old_deleted = Column('old_deleted', Boolean, default=False)
     old_deleted.create(table, populate_default=False)
@@ -517,14 +565,12 @@ def change_deleted_column_type_to_boolean(migrate_engine, table_name,
     table.c.deleted.drop()
     table.c.old_deleted.alter(name="deleted")
 
-    _restore_indexes_on_deleted_columns(migrate_engine, table_name, indexes)
+    _restore_indexes_on_deleted_columns(engine, table_name, indexes)
 
 
-def _change_deleted_column_type_to_boolean_sqlite(migrate_engine, table_name,
+def _change_deleted_column_type_to_boolean_sqlite(engine, table_name,
                                                   **col_name_col_instance):
-    insp = reflection.Inspector.from_engine(migrate_engine)
-    table = get_table(migrate_engine, table_name)
-
+    table = get_table(engine, table_name)
     columns = []
     for column in table.columns:
         column_copy = None
@@ -546,7 +592,7 @@ def _change_deleted_column_type_to_boolean_sqlite(migrate_engine, table_name,
     new_table.create()
 
     indexes = []
-    for index in insp.get_indexes(table_name):
+    for index in get_indexes(engine, table_name):
         column_names = [new_table.c[c] for c in index['column_names']]
         indexes.append(Index(index["name"], *column_names,
                              unique=index["unique"]))
@@ -558,12 +604,9 @@ def _change_deleted_column_type_to_boolean_sqlite(migrate_engine, table_name,
         else:
             c_select.append(table.c.deleted == table.c.id)
 
-    ins = InsertFromSelect(new_table, sqlalchemy.sql.select(c_select))
-    migrate_engine.execute(ins)
-
     table.drop()
     for index in indexes:
-        index.create(migrate_engine)
+        index.create(engine)
 
     new_table.rename(table_name)
     new_table.update().\
@@ -572,15 +615,13 @@ def _change_deleted_column_type_to_boolean_sqlite(migrate_engine, table_name,
         execute()
 
 
-def change_deleted_column_type_to_id_type(migrate_engine, table_name,
+def change_deleted_column_type_to_id_type(engine, table_name,
                                           **col_name_col_instance):
-    if migrate_engine.name == "sqlite":
+    if engine.name == "sqlite":
         return _change_deleted_column_type_to_id_type_sqlite(
-            migrate_engine, table_name, **col_name_col_instance)
-    insp = reflection.Inspector.from_engine(migrate_engine)
-    indexes = insp.get_indexes(table_name)
-
-    table = get_table(migrate_engine, table_name)
+            engine, table_name, **col_name_col_instance)
+    indexes = get_indexes(engine, table_name)
+    table = get_table(engine, table_name)
 
     new_deleted = Column('new_deleted', table.c.id.type,
                          default=_get_default_deleted_value(table))
@@ -594,10 +635,23 @@ def change_deleted_column_type_to_id_type(migrate_engine, table_name,
     table.c.deleted.drop()
     table.c.new_deleted.alter(name="deleted")
 
-    _restore_indexes_on_deleted_columns(migrate_engine, table_name, indexes)
+    _restore_indexes_on_deleted_columns(engine, table_name, indexes)
 
 
-def _change_deleted_column_type_to_id_type_sqlite(migrate_engine, table_name,
+def _is_deleted_column_constraint(constraint):
+    # NOTE(boris-42): There is no other way to check is CheckConstraint
+    #                 associated with deleted column.
+    if not isinstance(constraint, CheckConstraint):
+        return False
+    sqltext = str(constraint.sqltext)
+    # NOTE(zzzeek): SQLite never reflected CHECK contraints here
+    # in any case until version 1.1.   Safe to assume that any CHECK
+    # that's talking about the value of "deleted in (something)" is
+    # the boolean constraint we're looking to get rid of.
+    return bool(re.match(r".*deleted in \(.*\)", sqltext, re.I))
+
+
+def _change_deleted_column_type_to_id_type_sqlite(engine, table_name,
                                                   **col_name_col_instance):
     # NOTE(boris-42): sqlaclhemy-migrate can't drop column with check
     #                 constraints in sqlite DB and our `deleted` column has
@@ -608,8 +662,7 @@ def _change_deleted_column_type_to_id_type_sqlite(migrate_engine, table_name,
     #                 2) Copy all data from old to new table.
     #                 3) Drop old table.
     #                 4) Rename new table to old table name.
-    insp = reflection.Inspector.from_engine(migrate_engine)
-    meta = MetaData(bind=migrate_engine)
+    meta = MetaData(bind=engine)
     table = Table(table_name, meta, autoload=True)
     default_deleted_value = _get_default_deleted_value(table)
 
@@ -627,25 +680,9 @@ def _change_deleted_column_type_to_id_type_sqlite(migrate_engine, table_name,
                                  default=default_deleted_value)
         columns.append(column_copy)
 
-    def is_deleted_column_constraint(constraint):
-        # NOTE(boris-42): There is no other way to check is CheckConstraint
-        #                 associated with deleted column.
-        if not isinstance(constraint, CheckConstraint):
-            return False
-        sqltext = str(constraint.sqltext)
-        # NOTE(I159): in order to omit the CHECK constraint corresponding
-        # to `deleted` column we have to test these patterns which may
-        # vary depending on the SQLAlchemy version used.
-        constraint_markers = (
-            "deleted in (0, 1)",
-            "deleted IN (:deleted_1, :deleted_2)",
-            "deleted IN (:param_1, :param_2)"
-        )
-        return any(sqltext.endswith(marker) for marker in constraint_markers)
-
     constraints = []
     for constraint in table.constraints:
-        if not is_deleted_column_constraint(constraint):
+        if not _is_deleted_column_constraint(constraint):
             constraints.append(constraint.copy())
 
     new_table = Table(table_name + "__tmp__", meta,
@@ -653,17 +690,14 @@ def _change_deleted_column_type_to_id_type_sqlite(migrate_engine, table_name,
     new_table.create()
 
     indexes = []
-    for index in insp.get_indexes(table_name):
+    for index in get_indexes(engine, table_name):
         column_names = [new_table.c[c] for c in index['column_names']]
         indexes.append(Index(index["name"], *column_names,
                              unique=index["unique"]))
 
-    ins = InsertFromSelect(new_table, table.select())
-    migrate_engine.execute(ins)
-
     table.drop()
     for index in indexes:
-        index.create(migrate_engine)
+        index.create(engine)
 
     new_table.rename(table_name)
     deleted = True  # workaround for pyflakes
@@ -680,54 +714,6 @@ def _change_deleted_column_type_to_id_type_sqlite(migrate_engine, table_name,
         execute()
 
 
-def get_connect_string(backend, database, user=None, passwd=None,
-                       host='localhost'):
-    """Get database connection
-
-    Try to get a connection with a very specific set of values, if we get
-    these then we'll run the tests, otherwise they are skipped
-
-    DEPRECATED: this function is deprecated and will be removed from oslo.db
-    in a few releases. Please use the provisioning system for dealing
-    with URLs and database provisioning.
-
-    """
-    args = {'backend': backend,
-            'user': user,
-            'passwd': passwd,
-            'host': host,
-            'database': database}
-    if backend == 'sqlite':
-        template = '%(backend)s:///%(database)s'
-    else:
-        template = "%(backend)s://%(user)s:%(passwd)s@%(host)s/%(database)s"
-    return template % args
-
-
-def is_backend_avail(backend, database, user=None, passwd=None):
-    """Return True if the given backend is available.
-
-
-    DEPRECATED: this function is deprecated and will be removed from oslo.db
-    in a few releases. Please use the provisioning system to access
-    databases based on backend availability.
-
-    """
-    from oslo_db.sqlalchemy import provision
-
-    connect_uri = get_connect_string(backend=backend,
-                                     database=database,
-                                     user=user,
-                                     passwd=passwd)
-    try:
-        eng = provision.Backend._ensure_backend_available(connect_uri)
-        eng.dispose()
-    except exception.BackendNotAvailable:
-        return False
-    else:
-        return True
-
-
 def get_db_connection_info(conn_pieces):
     database = conn_pieces.path.strip('/')
     loc_pieces = conn_pieces.netloc.split('@')
@@ -742,29 +728,55 @@ def get_db_connection_info(conn_pieces):
     return (user, password, database, host)
 
 
-def index_exists(migrate_engine, table_name, index_name):
+def get_indexes(engine, table_name):
+    """Get all index list from a given table.
+
+    :param engine: sqlalchemy engine
+    :param table_name: name of the table
+    """
+
+    inspector = reflection.Inspector.from_engine(engine)
+    indexes = inspector.get_indexes(table_name)
+    return indexes
+
+
+def index_exists(engine, table_name, index_name):
     """Check if given index exists.
 
-    :param migrate_engine: sqlalchemy engine
-    :param table_name:     name of the table
-    :param index_name:     name of the index
+    :param engine:     sqlalchemy engine
+    :param table_name: name of the table
+    :param index_name: name of the index
     """
-    inspector = reflection.Inspector.from_engine(migrate_engine)
-    indexes = inspector.get_indexes(table_name)
+    indexes = get_indexes(engine, table_name)
     index_names = [index['name'] for index in indexes]
     return index_name in index_names
 
 
-def add_index(migrate_engine, table_name, index_name, idx_columns):
+def index_exists_on_columns(engine, table_name, columns):
+    """Check if an index on given columns exists.
+
+    :param engine: sqlalchemy engine
+    :param table_name: name of the table
+    :param columns: a list type of columns that will be checked
+    """
+    if not isinstance(columns, list):
+        columns = list(columns)
+    for index in get_indexes(engine, table_name):
+        if index['column_names'] == columns:
+            return True
+    return False
+
+
+def add_index(engine, table_name, index_name, idx_columns):
     """Create an index for given columns.
 
-    :param migrate_engine: sqlalchemy engine
-    :param table_name:     name of the table
-    :param index_name:     name of the index
-    :param idx_columns:    tuple with names of columns that will be indexed
+    :param engine:      sqlalchemy engine
+    :param table_name:  name of the table
+    :param index_name:  name of the index
+    :param idx_columns: tuple with names of columns that will be indexed
     """
-    table = get_table(migrate_engine, table_name)
-    if not index_exists(migrate_engine, table_name, index_name):
+    table = get_table(engine, table_name)
+    if not index_exists(engine, table_name, index_name):
         index = Index(
             index_name, *[getattr(table.c, col) for col in idx_columns]
         )
@@ -773,14 +785,14 @@ def add_index(migrate_engine, table_name, index_name, idx_columns):
         raise ValueError("Index '%s' already exists!" % index_name)
 
 
-def drop_index(migrate_engine, table_name, index_name):
+def drop_index(engine, table_name, index_name):
     """Drop index with given name.
 
-    :param migrate_engine: sqlalchemy engine
-    :param table_name:     name of the table
-    :param index_name:     name of the index
+    :param engine:     sqlalchemy engine
+    :param table_name: name of the table
+    :param index_name: name of the index
     """
-    table = get_table(migrate_engine, table_name)
+    table = get_table(engine, table_name)
     for index in table.indexes:
         if index.name == index_name:
             index.drop()
@@ -789,24 +801,24 @@ def drop_index(migrate_engine, table_name, index_name):
         raise ValueError("Index '%s' not found!" % index_name)
 
 
-def change_index_columns(migrate_engine, table_name, index_name, new_columns):
+def change_index_columns(engine, table_name, index_name, new_columns):
     """Change set of columns that are indexed by given index.
 
-    :param migrate_engine: sqlalchemy engine
-    :param table_name:     name of the table
-    :param index_name:     name of the index
-    :param new_columns:    tuple with names of columns that will be indexed
+    :param engine:      sqlalchemy engine
+    :param table_name:  name of the table
+    :param index_name:  name of the index
+    :param new_columns: tuple with names of columns that will be indexed
     """
-    drop_index(migrate_engine, table_name, index_name)
-    add_index(migrate_engine, table_name, index_name, new_columns)
+    drop_index(engine, table_name, index_name)
+    add_index(engine, table_name, index_name, new_columns)
 
 
 def column_exists(engine, table_name, column):
     """Check if table has given column.
 
-    :param engine:         sqlalchemy engine
-    :param table_name:     name of the table
-    :param column:         name of the colmn
+    :param engine:     sqlalchemy engine
+    :param table_name: name of the table
+    :param column:     name of the colmn
     """
     t = get_table(engine, table_name)
     return column in t.c
@@ -923,7 +935,7 @@ class DialectFunctionDispatcher(object):
     def _url_from_target(self, target):
         if isinstance(target, Connectable):
             return target.engine.url
-        elif isinstance(target, six.string_types):
+        elif isinstance(target, str):
             if "://" not in target:
                 target_url = sa_url.make_url("%s://" % target)
             else:
@@ -1021,11 +1033,10 @@ def get_non_innodb_tables(connectable, skip_tables=('migrate_version',
                                                     'alembic_version')):
     """Get a list of tables which don't use InnoDB storage engine.
 
-     :param connectable: a SQLAlchemy Engine or a Connection instance
-     :param skip_tables: a list of tables which might have a different
-                         storage engine
-     """
-
+    :param connectable: a SQLAlchemy Engine or a Connection instance
+    :param skip_tables: a list of tables which might have a different
+                        storage engine
+    """
     query_str = """
         SELECT table_name
         FROM information_schema.tables
@@ -1049,6 +1060,148 @@ def get_non_innodb_tables(connectable, skip_tables=('migrate_version',
     return [i[0] for i in noninnodb]
 
 
+def get_non_ndbcluster_tables(connectable, skip_tables=None):
+    """Get a list of tables which don't use MySQL Cluster (NDB) storage engine.
+
+    :param connectable: a SQLAlchemy Engine or Connection instance
+    :param skip_tables: a list of tables which might have a different
+                        storage engine
+    """
+    query_str = """
+        SELECT table_name
+        FROM information_schema.tables
+        WHERE table_schema = :database AND
+              engine != 'ndbcluster'
+    """
+
+    params = {}
+    if skip_tables:
+        params = dict(
+            ('skip_%s' % i, table_name)
+            for i, table_name in enumerate(skip_tables)
+        )
+
+        placeholders = ', '.join(':' + p for p in params)
+        query_str += ' AND table_name NOT IN (%s)' % placeholders
+
+    params['database'] = connectable.engine.url.database
+    query = text(query_str)
+    nonndbcluster = connectable.execute(query, **params)
+    return [i[0] for i in nonndbcluster]
+
+
+def get_foreign_key_constraint_name(engine, table_name, column_name):
+    """Find the name of foreign key in a table, given constrained column name.
+
+    :param engine: a SQLAlchemy engine (or connection)
+
+    :param table_name: name of table which contains the constraint
+
+    :param column_name: name of column that is constrained by the foreign key.
+
+    :return: the name of the first foreign key constraint which constrains
+     the given column in the given table.
+
+    """
+    insp = inspect(engine)
+    for fk in insp.get_foreign_keys(table_name):
+        if column_name in fk['constrained_columns']:
+            return fk['name']
+
+
+@contextlib.contextmanager
+def suspend_fk_constraints_for_col_alter(
+        engine, table_name, column_name, referents=[]):
+    """Detect foreign key constraints, drop, and recreate.
+
+    This is used to guard against a column ALTER that on some backends
+    cannot proceed unless foreign key constraints are not present.
+
+    e.g.::
+
+        from oslo_db.sqlalchemy.util import (
+            suspend_fk_constraints_for_col_alter
+        )
+
+        with suspend_fk_constraints_for_col_alter(
+            migrate_engine, "user_table",
+            referents=[
+                "local_user", "nonlocal_user", "project"
+            ]):
+            user_table.c.domain_id.alter(nullable=False)
+
+    :param engine: a SQLAlchemy engine (or connection)
+
+    :param table_name: target table name.  All foreign key constraints
+     that refer to the table_name / column_name will be dropped and recreated.
+
+    :param column_name: target column name.  all foreign key constraints
+     which refer to this column, either partially or fully, will be dropped
+     and recreated.
+
+    :param referents: sequence of string table names to search for foreign
+     key constraints.   A future version of this function may no longer
+     require this argument, however for the moment it is required.
+
+    """
+    if (
+        not ndb.ndb_status(engine)
+    ):
+        yield
+    else:
+        with engine.connect() as conn:
+            insp = inspect(conn)
+            fks = []
+            for ref_table_name in referents:
+                for fk in insp.get_foreign_keys(ref_table_name):
+                    if not fk.get('name'):
+                        raise AssertionError("foreign key hasn't a name.")
+                    if fk['referred_table'] == table_name and \
+                            column_name in fk['referred_columns']:
+                        fk['source_table'] = ref_table_name
+                        if 'options' not in fk:
+                            fk['options'] = {}
+                        fks.append(fk)
+
+            ctx = MigrationContext.configure(conn)
+            op = Operations(ctx)
+
+            for fk in fks:
+                op.drop_constraint(
+                    fk['name'], fk['source_table'], type_="foreignkey")
+            yield
+            for fk in fks:
+                op.create_foreign_key(
+                    fk['name'], fk['source_table'],
+                    fk['referred_table'],
+                    fk['constrained_columns'],
+                    fk['referred_columns'],
+                    onupdate=fk['options'].get('onupdate'),
+                    ondelete=fk['options'].get('ondelete'),
+                    deferrable=fk['options'].get('deferrable'),
+                    initially=fk['options'].get('initially'),
+                )
+
+
+def getargspec(fn):
+    """Inspects a function for its argspec.
+
+    This is to handle a difference between py2/3. The Python 2.x getargspec
+    call is deprecated in Python 3.x, with the suggestion to use the signature
+    call instead.
+
+    To keep compatibility with the results, while avoiding deprecation
+    warnings, this instead will use the getfullargspec instead.
+
+    :param fn: The function to inspect.
+    :returns: The argspec for the function.
+    """
+    if hasattr(pyinspect, 'getfullargspec'):
+        return pyinspect.getfullargspec(fn)
+
+    return pyinspect.getargspec(fn)
+
+
 class NonCommittingConnectable(object):
     """A ``Connectable`` substitute which rolls all operations back.
 
@@ -1068,12 +1221,16 @@ class NonCommittingConnectable(object):
 
     """
 
+    _nested_trans = None
+
     def __init__(self, connection):
         self.connection = connection
         self._trans = connection.begin()
         self._restart_nested()
 
     def _restart_nested(self):
+        if self._nested_trans is not None:
+            self._nested_trans.rollback()
         self._nested_trans = self.connection.begin_nested()
 
     def _dispose(self):
