@@ -18,6 +18,7 @@
 """Unit tests for SQLAlchemy specific code."""
 
 import logging
+import os
 
 import fixtures
 import mock
@@ -31,9 +32,11 @@ from sqlalchemy.ext.declarative import declarative_base
 
 from oslo_db import exception
 from oslo_db import options as db_options
+from oslo_db.sqlalchemy import enginefacade
+from oslo_db.sqlalchemy import engines
 from oslo_db.sqlalchemy import models
 from oslo_db.sqlalchemy import session
-from oslo_db.sqlalchemy import test_base
+from oslo_db.tests.sqlalchemy import base as test_base
 
 
 BASE = declarative_base()
@@ -48,7 +51,7 @@ class RegexpTable(BASE, models.ModelBase):
     bar = Column(String(255))
 
 
-class RegexpFilterTestCase(test_base.DbTestCase):
+class RegexpFilterTestCase(test_base._DbTestCase):
 
     def setUp(self):
         super(RegexpFilterTestCase, self).setUp()
@@ -62,8 +65,8 @@ class RegexpFilterTestCase(test_base.DbTestCase):
         self.addCleanup(test_table.drop)
 
     def _test_regexp_filter(self, regexp, expected):
-        _session = self.sessionmaker()
-        with _session.begin():
+        with enginefacade.writer.using(test_base.context):
+            _session = test_base.context.session
             for i in ['10', '20', u'♥']:
                 tbl = RegexpTable()
                 tbl.update({'bar': i})
@@ -71,7 +74,7 @@ class RegexpFilterTestCase(test_base.DbTestCase):
 
         regexp_op = RegexpTable.bar.op('REGEXP')(regexp)
         result = _session.query(RegexpTable).filter(regexp_op).all()
-        self.assertEqual([r.bar for r in result], expected)
+        self.assertEqual(expected, [r.bar for r in result])
 
     def test_regexp_filter(self):
         self._test_regexp_filter('10', ['10'])
@@ -86,7 +89,7 @@ class RegexpFilterTestCase(test_base.DbTestCase):
         self._test_regexp_filter(u'♦', [])
 
 
-class SQLiteSavepointTest(test_base.DbTestCase):
+class SQLiteSavepointTest(test_base._DbTestCase):
     def setUp(self):
         super(SQLiteSavepointTest, self).setUp()
         meta = MetaData()
@@ -209,17 +212,90 @@ class FakeDB2Engine(object):
         pass
 
 
-class MySQLDefaultModeTestCase(test_base.MySQLOpportunisticTestCase):
+class QueryParamTest(test_base.DbTestCase):
+    def _fixture(self):
+        from sqlalchemy import create_engine
+
+        def _mock_create_engine(*arg, **kw):
+            return create_engine("sqlite://")
+
+        return mock.patch(
+            "oslo_db.sqlalchemy.engines.sqlalchemy.create_engine",
+            side_effect=_mock_create_engine)
+
+    def test_add_assorted_params(self):
+        with self._fixture() as ce:
+            engines.create_engine(
+                "mysql+pymysql://foo:bar@bat",
+                connection_parameters="foo=bar&bat=hoho&bat=param2")
+
+        self.assertEqual(
+            ce.mock_calls[0][1][0].query,
+            {'bat': ['hoho', 'param2'], 'foo': 'bar'}
+        )
+
+    def test_add_no_params(self):
+        with self._fixture() as ce:
+            engines.create_engine(
+                "mysql+pymysql://foo:bar@bat")
+
+        self.assertEqual(
+            ce.mock_calls[0][1][0].query,
+            {}
+        )
+
+    def test_combine_params(self):
+        with self._fixture() as ce:
+            engines.create_engine(
+                "mysql+pymysql://foo:bar@bat/"
+                "?charset=utf8&param_file=tripleo.cnf",
+                connection_parameters="plugin=sqlalchemy_collectd&"
+                                      "collectd_host=127.0.0.1&"
+                                      "bind_host=192.168.1.5")
+
+        self.assertEqual(
+            ce.mock_calls[0][1][0].query,
+            {
+                'bind_host': '192.168.1.5',
+                'charset': 'utf8',
+                'collectd_host': '127.0.0.1',
+                'param_file': 'tripleo.cnf',
+                'plugin': 'sqlalchemy_collectd'
+            }
+        )
+
+    def test_combine_multi_params(self):
+        with self._fixture() as ce:
+            engines.create_engine(
+                "mysql+pymysql://foo:bar@bat/"
+                "?charset=utf8&param_file=tripleo.cnf&plugin=connmon",
+                connection_parameters="plugin=sqlalchemy_collectd&"
+                                      "collectd_host=127.0.0.1&"
+                                      "bind_host=192.168.1.5")
+
+        self.assertEqual(
+            ce.mock_calls[0][1][0].query,
+            {
+                'bind_host': '192.168.1.5',
+                'charset': 'utf8',
+                'collectd_host': '127.0.0.1',
+                'param_file': 'tripleo.cnf',
+                'plugin': ['connmon', 'sqlalchemy_collectd']
+            }
+        )
+
+
+class MySQLDefaultModeTestCase(test_base._MySQLOpportunisticTestCase):
     def test_default_is_traditional(self):
         with self.engine.connect() as conn:
             sql_mode = conn.execute(
                 "SHOW VARIABLES LIKE 'sql_mode'"
             ).first()[1]
 
-        self.assertTrue("TRADITIONAL" in sql_mode)
+        self.assertIn("TRADITIONAL", sql_mode)
 
 
-class MySQLModeTestCase(test_base.MySQLOpportunisticTestCase):
+class MySQLModeTestCase(test_base._MySQLOpportunisticTestCase):
 
     def __init__(self, *args, **kwargs):
         super(MySQLModeTestCase, self).__init__(*args, **kwargs)
@@ -312,8 +388,12 @@ class EngineFacadeTestCase(oslo_test.BaseTestCase):
         self.assertFalse(ses.autocommit)
         self.assertTrue(ses.expire_on_commit)
 
-    @mock.patch('oslo_db.sqlalchemy.session.get_maker')
-    @mock.patch('oslo_db.sqlalchemy.session.create_engine')
+    def test_direct_invocation_deprecated_args(self):
+        facade = session.EngineFacade("sqlite://", idle_timeout=59)
+        self.assertEqual(59, facade.get_engine().pool._recycle)
+
+    @mock.patch('oslo_db.sqlalchemy.orm.get_maker')
+    @mock.patch('oslo_db.sqlalchemy.engines.create_engine')
     def test_creation_from_config(self, create_engine, get_maker):
         conf = cfg.ConfigOpts()
         conf.register_opts(db_options.database_opts, group='database')
@@ -337,8 +417,9 @@ class EngineFacadeTestCase(oslo_test.BaseTestCase):
             connection_debug=100,
             max_pool_size=10,
             mysql_sql_mode='TRADITIONAL',
+            mysql_enable_ndb=False,
             sqlite_fk=False,
-            idle_timeout=mock.ANY,
+            connection_recycle_time=mock.ANY,
             retry_interval=mock.ANY,
             max_retries=mock.ANY,
             max_overflow=mock.ANY,
@@ -346,6 +427,10 @@ class EngineFacadeTestCase(oslo_test.BaseTestCase):
             sqlite_synchronous=mock.ANY,
             pool_timeout=mock.ANY,
             thread_checkin=mock.ANY,
+            json_serializer=None,
+            json_deserializer=None,
+            connection_parameters='',
+            logging_name=mock.ANY,
         )
         get_maker.assert_called_once_with(engine=create_engine(),
                                           autocommit=False,
@@ -397,15 +482,15 @@ class SQLiteConnectTest(oslo_test.BaseTestCase):
     def test_sqlite_fk_listener(self):
         engine = self._fixture(sqlite_fk=True)
         self.assertEqual(
-            engine.scalar("pragma foreign_keys"),
-            1
+            1,
+            engine.scalar("pragma foreign_keys")
         )
 
         engine = self._fixture(sqlite_fk=False)
 
         self.assertEqual(
-            engine.scalar("pragma foreign_keys"),
-            0
+            0,
+            engine.scalar("pragma foreign_keys")
         )
 
     def test_sqlite_synchronous_listener(self):
@@ -414,31 +499,31 @@ class SQLiteConnectTest(oslo_test.BaseTestCase):
         # "The default setting is synchronous=FULL." (e.g. 2)
         # http://www.sqlite.org/pragma.html#pragma_synchronous
         self.assertEqual(
-            engine.scalar("pragma synchronous"),
-            2
+            2,
+            engine.scalar("pragma synchronous")
         )
 
         engine = self._fixture(sqlite_synchronous=False)
 
         self.assertEqual(
-            engine.scalar("pragma synchronous"),
-            0
+            0,
+            engine.scalar("pragma synchronous")
         )
 
 
-class MysqlConnectTest(test_base.MySQLOpportunisticTestCase):
+class MysqlConnectTest(test_base._MySQLOpportunisticTestCase):
 
     def _fixture(self, sql_mode):
         return session.create_engine(self.engine.url, mysql_sql_mode=sql_mode)
 
     def _assert_sql_mode(self, engine, sql_mode_present, sql_mode_non_present):
         mode = engine.execute("SHOW VARIABLES LIKE 'sql_mode'").fetchone()[1]
-        self.assertTrue(
-            sql_mode_present in mode
+        self.assertIn(
+            sql_mode_present, mode
         )
         if sql_mode_non_present:
-            self.assertTrue(
-                sql_mode_non_present not in mode
+            self.assertNotIn(
+                sql_mode_non_present, mode
             )
 
     def test_set_mode_traditional(self):
@@ -495,7 +580,7 @@ class MysqlConnectTest(test_base.MySQLOpportunisticTestCase):
                 )
             )
         ):
-            session._init_events.dispatch_on_drivername("mysql")(test_engine)
+            engines._init_events.dispatch_on_drivername("mysql")(test_engine)
 
             test_engine.raw_connection()
         self.assertIn('Unable to detect effective SQL mode',
@@ -556,99 +641,151 @@ class CreateEngineTest(oslo_test.BaseTestCase):
         self.args = {'connect_args': {}}
 
     def test_queuepool_args(self):
-        session._init_connection_args(
-            url.make_url("mysql://u:p@host/test"), self.args,
+        engines._init_connection_args(
+            url.make_url("mysql+pymysql://u:p@host/test"), self.args,
             max_pool_size=10, max_overflow=10)
-        self.assertEqual(self.args['pool_size'], 10)
-        self.assertEqual(self.args['max_overflow'], 10)
+        self.assertEqual(10, self.args['pool_size'])
+        self.assertEqual(10, self.args['max_overflow'])
 
     def test_sqlite_memory_pool_args(self):
         for _url in ("sqlite://", "sqlite:///:memory:"):
-            session._init_connection_args(
+            engines._init_connection_args(
                 url.make_url(_url), self.args,
                 max_pool_size=10, max_overflow=10)
 
             # queuepool arguments are not peresnet
-            self.assertTrue(
-                'pool_size' not in self.args)
-            self.assertTrue(
-                'max_overflow' not in self.args)
+            self.assertNotIn(
+                'pool_size', self.args)
+            self.assertNotIn(
+                'max_overflow', self.args)
 
-            self.assertEqual(self.args['connect_args']['check_same_thread'],
-                             False)
+            self.assertEqual(False,
+                             self.args['connect_args']['check_same_thread'])
 
             # due to memory connection
-            self.assertTrue('poolclass' in self.args)
+            self.assertIn('poolclass', self.args)
 
     def test_sqlite_file_pool_args(self):
-        session._init_connection_args(
+        engines._init_connection_args(
             url.make_url("sqlite:///somefile.db"), self.args,
             max_pool_size=10, max_overflow=10)
 
         # queuepool arguments are not peresnet
-        self.assertTrue('pool_size' not in self.args)
-        self.assertTrue(
-            'max_overflow' not in self.args)
+        self.assertNotIn('pool_size', self.args)
+        self.assertNotIn(
+            'max_overflow', self.args)
 
         self.assertFalse(self.args['connect_args'])
 
         # NullPool is the default for file based connections,
         # no need to specify this
-        self.assertTrue('poolclass' not in self.args)
+        self.assertNotIn('poolclass', self.args)
+
+    def _test_mysql_connect_args_default(self, connect_args):
+        self.assertEqual({'charset': 'utf8', 'use_unicode': 1},
+                         connect_args)
 
     def test_mysql_connect_args_default(self):
-        session._init_connection_args(
+        engines._init_connection_args(
             url.make_url("mysql://u:p@host/test"), self.args)
-        self.assertEqual(self.args['connect_args'],
-                         {'charset': 'utf8', 'use_unicode': 0})
+        self._test_mysql_connect_args_default(self.args['connect_args'])
 
     def test_mysql_oursql_connect_args_default(self):
-        session._init_connection_args(
+        engines._init_connection_args(
             url.make_url("mysql+oursql://u:p@host/test"), self.args)
-        self.assertEqual(self.args['connect_args'],
-                         {'charset': 'utf8', 'use_unicode': 0})
+        self._test_mysql_connect_args_default(self.args['connect_args'])
+
+    def test_mysql_pymysql_connect_args_default(self):
+        engines._init_connection_args(
+            url.make_url("mysql+pymysql://u:p@host/test"), self.args)
+        self.assertEqual({'charset': 'utf8'}, self.args['connect_args'])
 
     def test_mysql_mysqldb_connect_args_default(self):
-        session._init_connection_args(
+        engines._init_connection_args(
             url.make_url("mysql+mysqldb://u:p@host/test"), self.args)
-        self.assertEqual(self.args['connect_args'],
-                         {'charset': 'utf8', 'use_unicode': 0})
+        self._test_mysql_connect_args_default(self.args['connect_args'])
 
     def test_postgresql_connect_args_default(self):
-        session._init_connection_args(
+        engines._init_connection_args(
             url.make_url("postgresql://u:p@host/test"), self.args)
-        self.assertEqual(self.args['client_encoding'], 'utf8')
+        self.assertEqual('utf8', self.args['client_encoding'])
         self.assertFalse(self.args['connect_args'])
 
     def test_mysqlconnector_raise_on_warnings_default(self):
-        session._init_connection_args(
+        engines._init_connection_args(
             url.make_url("mysql+mysqlconnector://u:p@host/test"),
             self.args)
-        self.assertEqual(self.args['connect_args']['raise_on_warnings'], False)
+        self.assertEqual(False, self.args['connect_args']['raise_on_warnings'])
 
     def test_mysqlconnector_raise_on_warnings_override(self):
-        session._init_connection_args(
+        engines._init_connection_args(
             url.make_url(
                 "mysql+mysqlconnector://u:p@host/test"
                 "?raise_on_warnings=true"),
             self.args
         )
 
-        self.assertFalse('raise_on_warnings' in self.args['connect_args'])
+        self.assertNotIn('raise_on_warnings', self.args['connect_args'])
 
     def test_thread_checkin(self):
         with mock.patch("sqlalchemy.event.listens_for"):
             with mock.patch("sqlalchemy.event.listen") as listen_evt:
-                session._init_events.dispatch_on_drivername(
+                engines._init_events.dispatch_on_drivername(
                     "sqlite")(mock.Mock())
 
         self.assertEqual(
             listen_evt.mock_calls[0][1][-1],
-            session._thread_yield
+            engines._thread_yield
+        )
+
+    def test_warn_on_missing_driver(self):
+
+        warnings = mock.Mock()
+
+        def warn_interpolate(msg, args):
+            # test the interpolation itself to ensure the password
+            # is concealed
+            warnings.warning(msg % args)
+
+        with mock.patch(
+                "oslo_db.sqlalchemy.engines.LOG.warning",
+                warn_interpolate):
+
+            engines._vet_url(
+                url.make_url("mysql://scott:tiger@some_host/some_db"))
+            engines._vet_url(url.make_url(
+                "mysql+mysqldb://scott:tiger@some_host/some_db"))
+            engines._vet_url(url.make_url(
+                "mysql+pymysql://scott:tiger@some_host/some_db"))
+            engines._vet_url(url.make_url(
+                "postgresql+psycopg2://scott:tiger@some_host/some_db"))
+            engines._vet_url(url.make_url(
+                "postgresql://scott:tiger@some_host/some_db"))
+
+        self.assertEqual(
+            [
+                mock.call.warning(
+                    "URL mysql://scott:***@some_host/some_db does not contain "
+                    "a '+drivername' portion, "
+                    "and will make use of a default driver.  "
+                    "A full dbname+drivername:// protocol is recommended. "
+                    "For MySQL, it is strongly recommended that "
+                    "mysql+pymysql:// "
+                    "be specified for maximum service compatibility",
+
+                ),
+                mock.call.warning(
+                    "URL postgresql://scott:***@some_host/some_db does not "
+                    "contain a '+drivername' portion, "
+                    "and will make use of a default driver.  "
+                    "A full dbname+drivername:// protocol is recommended."
+                )
+            ],
+            warnings.mock_calls
         )
 
 
-class ProcessGuardTest(test_base.DbTestCase):
+class ProcessGuardTest(test_base._DbTestCase):
     def test_process_guard(self):
         self.engine.dispose()
 
@@ -676,7 +813,7 @@ class ProcessGuardTest(test_base.DbTestCase):
         self.assertEqual(new_dbapi_id, newer_dbapi_id)
 
 
-class PatchStacktraceTest(test_base.DbTestCase):
+class PatchStacktraceTest(test_base._DbTestCase):
 
     def test_trace(self):
         engine = self.engine
@@ -693,7 +830,7 @@ class PatchStacktraceTest(test_base.DbTestCase):
 
         with mock.patch("traceback.extract_stack", side_effect=extract_stack):
 
-            session._add_trace_comments(engine)
+            engines._add_trace_comments(engine)
             conn = engine.connect()
             orig_do_exec = engine.dialect.do_execute
             with mock.patch.object(engine.dialect, "do_execute") as mock_exec:
@@ -704,4 +841,5 @@ class PatchStacktraceTest(test_base.DbTestCase):
             call = mock_exec.mock_calls[0]
 
             # we're the caller, see that we're in there
-            self.assertIn("tests/sqlalchemy/test_sqlalchemy.py", call[1][1])
+            caller = os.path.join("tests", "sqlalchemy", "test_sqlalchemy.py")
+            self.assertIn(caller, call[1][1])
